import copy
import inspect
import itertools
import operator
import re
import shutil
import string
from abc import ABC, abstractmethod
from importlib import import_module
from pathlib import Path
from typing import Callable, Literal, Optional, Union, cast
import datetime
import glayout.syntaxer.dynamic_load

import nltk
import glayout.syntaxer.nltk_init_deps

def list_cartesian_product(list1: list, list2: list, both: bool=False) -> list:
    """Compute the Cartesian product of two lists and combine elements into a list of strings.

    Args:
        list1 (list): The first list.
        list2 (list): The second list.
        both (bool): provide the cartesian product of list1,list2 AND list2,list1

    Returns:
        list: A list containing the Cartesian product of the input lists as strings.
    """
    # Cartesian product of the two lists
    cartesian_product = list(itertools.product(list1, list2))
    # Combine the elements into a list of strings
    combined_list = [' '.join(map(str, pair)) for pair in cartesian_product]
    return combined_list + (list_cartesian_product(list2,list1,False) if both else [])


# used to automatically update supported actions whenever a class which inherits from GlayoutAction is added to this file
def get_all_derived_classes(class_type, remove_parent: bool=True) -> list:
    """Finds all derived classes of a particular parent type in the current python file\
    Args:
        class_type (any): the parent type you want derived classes for
        remove_parent (bool): if True, do not include the parent class in the returned list
    Returns list of derived classes
    """
    all_classes = [classname for name, classname in globals().items() if isinstance(classname, type)]
    glayout_actions = [classname for classname in all_classes if issubclass(classname, GlayoutAction)]
    try:
        glayout_actions.remove(GlayoutAction)
    except ValueError:
        pass
    return glayout_actions


def parse_direction(direction: str) -> Union[str,None]:
    """parses a string to see if it contains a valid direction are up/north/above, right/east, left/west, down/south/below
    Args: 
        direction (str)
    Returns None if no direction found or an integer 0-3 (0=west,1=north,2=south,3=east)
    """
    direction = direction.lower().strip()
    findany = lambda strlst : any(ind in direction for ind in strlst)
    if findany(["left","west"]):
        return 0
    elif findany(["up","north","above"]):
        return 1
    elif findany(["right","east"]):
        return 2
    elif findany(["down","south","below"]):
        return 3
    else:
        return None



class ParametersList:
    """store function parameter information
    self.params is a list with the following format:
        [ {name, defaultvalue, type}, ... ]
    if niether type nor default is specified, both fields contain inspect.Parameter.empty
    self.func_name: str the name of the function
    """
    class noprintobj:
        def __init__(self,data):
            self.data = data
        def __str__(self):
            return str(self.data)
        def __repr__(self):
            return str(self.data)
    
    def __init__(self, func)->dict:
        self.func_name = func.__name__
        parameters = inspect.signature(func).parameters
        self.params = list()
        for param_name, param in parameters.items():
            paraminfo = {"name":param_name, 'defaultvalue':param.default, 'type':param.annotation}
            self.params.append(paraminfo)

    def __iter__(self):
        return iter(self.params)
    def names(self):
        return [param["name"] for param in self.params]
    def find(self, param_name: str):
        """look for param with name param_name. 
        If found, return info, else return None"""
        for param in self.params:
            if param["name"]==param_name:
                return param
        return None
    def __str__(self):
        return str(self.params)
    
    def __construct_glayers(self) -> dict:
        glayerids = ['met1', 'met2', 'met3', 'met4', 'met5', 'metal1', 'metal2', 'metal3', 'metal4', 'metal5', 'capmet', 'active', 'diffusion', 'tap', 'welltap', 'n+s/d', 'p+s/d', 'polysilicon', 'poly', 'p+', 'n+', 'deep nwell', 'pwell', 'nwell', 'dnwell']
        glayerids += ["metal 1", "metal 2","metal 3","metal 4","metal 5","met 1","met 2","met 3","met 4","met 5"]
        glayerids += ["via 1", "via 2","via 3","via 4","via1","via2","via3","via4","metal contact","mcon","metal con"]
        glayers = dict()
        for glayerid in glayerids:
            if "p+" in glayerid:
                glayers[glayerid] = "p+s/d"
            elif "n+" in glayerid:
                glayers[glayerid] = "n+s/d"
            elif "poly" in glayerid:
                glayers[glayerid] = "poly"
            elif "cap" in glayerid:
                glayers[glayerid] = "capmet"
            elif "tap" in glayerid:
                glayers[glayerid] = "active_tap"
            elif "active" in glayerid or "diff" in glayerid:
                glayers[glayerid] = "active_diff"
            elif "well" in glayerid:
                if "d" in glayerid:
                    glayers[glayerid] = "dnwell"
                elif "n" in glayerid:
                    glayers[glayerid] = "nwell"
                else:
                    glayers[glayerid] = "pwell"
            elif "via" in glayerid:
                for i in range(1,5):
                    if str(i) in glayerid:
                        glayers[glayerid] = f"via{i}"
            elif "met" in glayerid:
                for i in range(1,6):
                    if str(i) in glayerid:
                        glayers[glayerid] = f"met{i}"
            elif "con" in glayerid:
                glayers[glayerid] = "mcon"
            else:
                raise ValueError("could not parse this id")
        return glayers
    def __cfgformat_from_list(self, strlist: list[str]):
        rtrstr = str()
        for ele in strlist:
            rtrstr += "\'" + ele + "\' | "
        rtrstr =rtrstr.removesuffix("| ")
        return rtrstr
    def __custom_tokenize(self, sentence: str) -> list:
        sentence = sentence.strip().removesuffix(".").removeprefix("with")
        # save and replace all instances of strings with indicator words
        string_pattern = re.compile(r'\'[^\']*\'|\"[^\"]*\"')
        string_list = [str(match.group()) for match in string_pattern.finditer(sentence)]
        string_list = [w.replace("\"","").replace("'","").strip() for w in string_list]
        sentence = re.sub(string_pattern, "sTRingindICatorWOrd", sentence)
        # add a space after every symbol (except period and _)
        rsymbols = list(set(string.punctuation) - {'.', '_'})
        parsedsentence = str()
        for char in sentence:
            if char in rsymbols:
                parsedsentence += f" {char} "
            else:
                parsedsentence += char
        sentence = copy.deepcopy(parsedsentence)
        # double space (add one space everywhere there is a space)
        parsedsentence = str()
        for char in sentence:
            if char.isspace():
                parsedsentence += f" {char}"
            else:
                parsedsentence += char
        sentence = copy.deepcopy(parsedsentence)
        # save and replace all instances of numbers with indicator words
        number_pattern = re.compile(r'(?:^|\s|\b)-?\d+(\.\d+)?([eE]-?\d+)?(?=\s|$|[^\w\d])')
        number_list = [str(match.group()) for match in number_pattern.finditer(sentence)]
        number_list = [num.strip() for num in number_list]
        sentence = re.sub(number_pattern, "nUMptindICatorWOrd", sentence)
        # save numbers and strings to list and return tokenized sentence
        self.__number_list = number_list
        self.__string_list = string_list
        return sentence.split()
    def __get_str_grammar(self, known_paramsORvars: list) -> str:
        if len(known_paramsORvars)==0:
            known_paramsORvars = ["thiswordwillsurelyneverevercomeupduringparsing"]
        basic_grammar = str()
        basic_grammar += """S -> ArgList\n"""
        basic_grammar += """ArgList -> ArgDesc | ArgDesc ArgList | ArgDesc ListSeparator ArgList\n"""
        basic_grammar += """ListSeparator -> ',' | ', and' | 'and'\n"""
        basic_grammar += """ArgDesc -> Value Paramname | Paramname Value | Paramname Filler Value | Value Filler Paramname\n"""
        basic_grammar += """Filler -> 'is' | '=' | 'equal' | 'equals' | 'of' | 'for'\n"""
        basic_grammar += """Value  -> Number | Glayer | Dictionary | Boolean | List | Tuple | String | paramORvar\n"""
        basic_grammar += """Vpair -> Value ',' Value | Value Value \n"""
        basic_grammar += """Values -> Vpair | Vpair Values | Vpair ',' Values\n"""
        basic_grammar += """Number -> 'nUMptindICatorWOrd'\n"""
        basic_grammar += """String -> 'sTRingindICatorWOrd'\n"""
        basic_grammar += """Parity -> '+' | '-' | 'positive' | 'negative' | 'plus' | 'minus'\n"""
        basic_grammar += """Boolean -> 'true' | 'false' | 'True' | 'False'\n"""
        basic_grammar += """List   -> '[' Values ']' | '[' Value ']'\n"""
        basic_grammar += """Tuple  -> '(' Values ')' | '(' Value ')' | Values\n"""
        basic_grammar += """Dictionary  -> '{' keyvalpairs '}'\n"""
        basic_grammar += f"""paramORvar  -> {self.__cfgformat_from_list(known_paramsORvars)}\n"""
        basic_grammar += """keyvalpairs -> Value ':' Value | Value ':' Value ',' keyvalpairs\n"""
        basic_grammar += f"Glayer -> {self.__cfgformat_from_list(self.__construct_glayers().keys())}\n"
        basic_grammar += f"Paramname -> {self.__cfgformat_from_list(self.names())}\n"
        return basic_grammar

    def parse_user_params(self, known_paramsORvars: list, user_input_params: Optional[str]=None):
        """Take user params as a string and parse+error check to produce a dictionary of params
        Args:
            user_input_params (str)
        """
        if user_input_params is None or user_input_params.isspace() or len(user_input_params)==0:
            return {}
        # tokenize
        tokens = self.__custom_tokenize(user_input_params)
        # try to parse syntax
        # formulate the grammar using the parameters
        self.__grammar = nltk.CFG.fromstring(self.__get_str_grammar(known_paramsORvars))
        self.__grammar_parser = nltk.ChartParser(self.__grammar)
        try:
            mytree = list(self.__grammar_parser.parse(tokens))[0]
        except Exception as serr:
            raise ValueError("You may have provided a parameter that does not exist for this function OR there may be some other error") from serr
        # reinsert numbers and strings
        numlist_index = 0
        strlist_index = 0
        for pos in mytree.treepositions('leaves'):
            if "nUMptindICatorWOrd" in mytree[pos]:
                mytree[pos] = self.__number_list[numlist_index]
                numlist_index += 1
            elif "sTRingindICatorWOrd" in mytree[pos]:
                mytree[pos] = "\""+self.__string_list[strlist_index]+"\""
                strlist_index += 1
        # grab user params from the tree
        user_params = dict()
        for i, ArgDesc in enumerate(mytree.subtrees(lambda t: t.label()=="ArgDesc")):
            # TODO: implement param name aliasing
            paramname = "".join(list(ArgDesc.subtrees(lambda A: A.label()=="Paramname"))[0].leaves())
            # TODO: implement more sophisticated value parsing, and properly add double qoutes are glayers
            value = " ".join(list(ArgDesc.subtrees(lambda A: A.label()=="Value"))[0].leaves())
            user_params[paramname] = self.noprintobj(value)
        # split off kwargs into actual key word arguments in the params dictionary
        if user_params.get("kwargs") is not None:
            # TODO: maybe does not support edgecase where kwargs contains a dictionary? check
            user_params[self.noprintobj("None")] = self.noprintobj("None,"+user_params["kwargs"].data.split("{")[-1].replace("}",""))
            del user_params["kwargs"]
        user_params = str(user_params).replace("None: None,","")
        return user_params




######################################################
# Data structures for possible operations with Glayout
######################################################

class GlayoutAction(ABC):    
    """each GlayoutAction corresponds to one or several Glayout code operations
    GlayoutActions are convertible to code via the "get_code" method

    Args:
        ABC (_type_): _description_
    """
    # GlayoutActions must be convertible to Glayout code via the get_code method
    @abstractmethod
    def get_code(self) -> str:
        pass
    
    # GlayoutActions must have a test_suite to confirm functionality
    @classmethod
    @abstractmethod
    def test(cls):
        pass


class ImportCell(GlayoutAction):
    """Import an existing CellFactory
    The following information is stored:
        identifiers: list[str]
        handle: callable
        relative_path: str
        parameters: ParametersList
    """
    #@validate_call
    def __init__(self, component_identifiers: list[str], component_name: str, module_path: Optional[Union[str, Path]]=None, overwrite_module: bool=True):
        """Store neccessary information 
        Will try to import the module; if it fails, will raise an exception
        Args:
            component_identifiers (list[str]): a list of potential names the user may use to identify this component
            component_name (str): name of the thing we want to import from the target python module
            module_path (str | Path | None): path of the module we want to import from
            relative_path (bool): if true, takes the module_path directly as a python relative import path
        """
        # try to find an os path to the desired module by looking in Glayout
        if module_path is None:
            # try to resolve path of glayout package
            matching_files = list(Path("./").resolve().rglob("glayout"))
            if len(matching_files)>0:
                glayout_path = (matching_files[0].resolve() / "flow").resolve()
            else:
                raise FileNotFoundError("Glayout.flow package not found in a sub folder of ../ directory")
            # try to resolve path of user module in Glayout package
            # this will look for the file which starts with "component_name" ANYWHERE in the glayout.flow package
            matching_files = list(glayout_path.rglob(str(component_name)+".py"))
            matching_files += list(glayout_path.rglob(str(component_name)+".convo"))
            # also check the test_cases directory for convo files
            matching_files += list((glayout_path / "../llm/syntax_data/convos").rglob(str(component_name)+".convo"))
            if len(matching_files)>0:
                module_path = matching_files[-1].resolve()
            else:
                raise FileNotFoundError("Could not find a module called "+str(component_name)+" in Glayout")
        # copy convo file into glayout Components and then create a python file and copy it into glayout components
        if str(module_path).endswith(".convo"):
            components_directory = glayout_path / "components"
            if module_path.parent != components_directory:
                shutil.copy(module_path, components_directory)
            glayoutcode = glayout.syntaxer.dynamic_load.run_session(module_path,True)
            component_name = glayout.syntaxer.dynamic_load.get_funccell_name(glayoutcode)
            module_path = components_directory / (component_name + ".py")
            # write python code from a convo to the components directory
            with open(module_path,mode="w") as pycell:
                pycell.write(glayoutcode)
        # resolve module_path into a python relative import path (if it is not already so)
        if any(ind in Path(module_path).as_posix() for ind in ["/",".py"]):
            module_path = Path(module_path).resolve()
            # error check that the path is indeed a python module and a real file
            if module_path.suffix != ".py" or (not module_path.is_file()):
                raise ValueError("module to import must be a python file ending in .py")
            # if the module is not in the glayout.flow directory, it should be copied into current directory so it can be imported
            module_path_parts = [part for part in module_path.parts]
            if "flow" in module_path_parts:
                module_path = "glayout." + ".".join(module_path_parts[module_path_parts.index("flow"):]).removesuffix(".py")
            else:
                # ensure the file does not already exist in the current directory, then copy
                if Path("./"+module_path.name).is_file() and not overwrite_module:
                    raise ValueError("module name cannot conflict with existing module name in "+str(Path("./").resolve()))
                shutil.copy(module_path,"./")
                module_path = module_path.stem
        # import callable (and make sure what we are importing is indeed callable)
        python_import_path = str(module_path)
        imported_obj = self.from_module_import(python_import_path, component_name)
        if not callable(imported_obj):
            raise TypeError("the thing you are trying to import must be callable")
        # store relavent import information
        self.identifiers = component_identifiers
        self.handle = imported_obj
        self.relative_path = python_import_path
        self.parameters = ParametersList(imported_obj)
    
    @classmethod
    def from_module_import(cls, module: str, nameofobjtoimport: str):
        """this function is similar to the following line of python
        from {module} import {nameofobjtoimport}
        Returns: the object that was imported
        """
        module_object = import_module(module)
        return getattr(module_object, nameofobjtoimport)
    
    def get_code(self) -> str:
        """returns as str a whole import line of code"""
        return "from "+self.relative_path+" import "+self.handle.__name__
    
    @classmethod
    def test(cls):
        tests = list()
        tests.append([["nmos","nfet"],"nmos","glayout.flow.primitives.fet"])
        tests.append([["pmos","pfet"],"pmos","glayout.flow.primitives.fet"])
        tests.append([["guardring","tapring","welltap","well tap", "tap ring"],"tapring","glayout.flow.primitives.guardring"])
        tests.append([["mimcap"],"mimcap","glayout.flow.primitives.mimcap"])
        tests.append([["mimcap array","mimcaparray","mimcap_array"],"mimcap_array","glayout.flow.primitives.mimcap"])
        tests.append([["via","via stack","via_stack"],"via_stack","glayout.flow.primitives.via_gen"])
        tests.append([["via array","via_array"],"via_array","glayout.flow.primitives.via_gen"])
        tests.append([["nlp"],"test_stream","../interpreter/deprecated/../deprecated/practice_stream.py"])
        tests.append([["nmos","nfet"],"nmos","./glayout/primitives/fet.py"])
        #print("expected\t\t\tresult")
        for testinst in tests:
            print(ImportCell(*testinst).get_code())
    

class CreateCellParameter(GlayoutAction):
    """Create a new cell parameter
    The following information is stored:
        varname: str
        type: Any
        defaultvalue: int/float/None
        reqs: tuple[opcode, value]
    """
    #@validate_call
    def __init__(
            self, 
            varname: str, 
            vartype: Optional[Literal[int, float]]=None, 
            defaultvalue: Optional[Union[int,float]]=None, 
            requirements: Optional[tuple[Literal[">",">=","<","<=","!=","=="],Union[float,int]]]=None
        ):
        """Create a cell parameter
        Args:
            varname (str): name of parameter
            vartype (Literal[int, float]): type of parameter is either int or float
            defaultvalue: default value for parameter, Optional, None indicates no default
            requirements: tuple[condition,value] where variable must be {condition} {value}
            value can be a value or it can refer to another parameter
                for example, variable must be <= 5
        """
        self.varname = str(varname)
        self.type = vartype
        self.defaultvalue = defaultvalue if self.type is None or defaultvalue is None else self.type(defaultvalue)
        self.reqs = None
        if requirements is not None:
            required_op = None
            if requirements[0]==">":
                required_op = operator.gt
            elif requirements[0]=="<":
                required_op = operator.lt
            elif requirements[0]==">=":
                required_op = operator.ge
            elif requirements[0]=="<=":
                required_op = operator.le
            elif requirements[0]=="==":
                required_op = operator.eq
            elif requirements[0]=="!=":
                required_op = operator.ne
            self.reqs=(required_op,requirements[1])
    
    def get_code(self) -> str:
        """returns as str a single parameter with type annotation to go in function definition
        includes a tab in front and comma at the end
        """
        param = self.varname
        if self.type is not None:
            param += ": "+self.type.__name__
        if self.defaultvalue is not None:
            param += "="+str(self.defaultvalue)
        if self.reqs is not None:
            raise NotImplementedError("parameter error checking has not yet been implemented")
        return "\t" + param + ", "
    
    def __str__(self) -> str:
        return self.type.__name__ + " " + self.varname + ((" = "+str(self.defaultvalue)) if self.defaultvalue is not None else "")
    
    @classmethod
    def test(cls):
        tests = list()
        tests.append(["width",float])
        tests.append(["length",float,0])
        tests.append(["fingers",int,3])
        for testinst in tests:
            print(CreateCellParameter(*testinst).get_code())


class CreateWorkingVariable(GlayoutAction):
    """create a new working variable
    The following information is stored:
        varname: str
        expression: str
    """
    #@validate_call
    def __init__(self, varname: str, input_expression: str):
        """create a working variable
        Args:
            varname (str): name of working variable
            input_expression (str): mathematical expression or value
        """
        self.varname = str(varname)
        self.expression = input_expression #sympify(input_expression)
        #self.handle = symbols(self.varname)
    
    def get_code(self):
        """return a complete line of code that defines the working variable"""
        return self.varname + " = "+str(self.expression)
    
    @classmethod
    def test(cls):
        tests = list()
        tests.append(["max_metal_sep","pdk.util_max_metal_sep()"])
        tests.append(["pi","3.1415"])
        for testinst in tests:
            print(CreateWorkingVariable(*testinst).get_code())


class PlaceCell(GlayoutAction):
    """Place an existing cell at the origin
    The following information is stored:
        name: str (name for the component reference + _reference)
        handle: Callable
        params: parameters to pass to callable
    """
    ref_suffix = "ref"
    #@validate_call
    def __init__(self, toplvl_name: str, generator_handle: Callable, component_name: str, user_input_parameters: str, known_paramsORvars: list):
        """Store all information neccessary to place a cell without moving it anywhere
        Args:
            generator_handle (Callable): cell factory
            component_name (str): name to assign to this component
            user_input_parameters (str): just the portion of user input containing the parameters for the component
            user input should be parameter name followed by value (optionally with a filler word e.g. [of, =, is])
            user input should contain NOTHING except parameters and values
            known_paramsORvars (list): a list containing names of the currently defined parameters or variables
        """
        # append to the place table
        self.name = component_name
        self.handle = generator_handle
        self.params = ParametersList(generator_handle).parse_user_params(known_paramsORvars, user_input_parameters)
        self.toplvl_name = toplvl_name
    
    def get_code(self) -> str:
        """return 3 complete lines of code for placing a component"""
        comment = f"# placing {self.name} centered at the origin"
        l1 = f"{self.name} = {self.handle.__name__}(pdk,**{str(self.params)})"
        l2 = f"{self.name}_{self.ref_suffix} = prec_ref_center({self.name})"
        l3 = f"{self.toplvl_name}.add({self.name}_{self.ref_suffix})"
        l4 = f"{self.toplvl_name}.add_ports({self.name}_{self.ref_suffix}.get_ports_list(),prefix=\"{self.name}_\")"
        return comment + "\n" + l1 + "\n" + l2 + "\n" + l3 + "\n" + l4
    
    @classmethod
    def test(cls):
        exp_name = "example_toplvl"
        tests = list()
        from glayout.flow.primitives.fet import nmos
        tests.append([exp_name,nmos,"mirror","width of 4 length of 1"])
        for testinst in tests:
            print(PlaceCell.get_code(*testinst))


class AbsoluteMove(GlayoutAction):
    """Move an existing Component by an absolute (x,y) distance
    The following information is stored:
        name: str (name of component to move should match the place table)
        move_distance: tuple[float,float] (x,y) absolute move distance
    """
    #@validate_call
    def __init__(self, name_of_component_to_move: str, toplvl_name:str, absolute_move_info: tuple[float,float]):
        self.name = str(name_of_component_to_move)
        self.move_distance = eval(absolute_move_info) if isinstance(absolute_move_info,str) else absolute_move_info
        self.toplvl_name = toplvl_name
    
    def get_code(self) -> str:
        xmov = self.move_distance[0]
        ymov = self.move_distance[1]
        movecode = f"{self.name}_ref.movex({xmov}).movey({ymov})\n"
        movecode += f"remove_ports_with_prefix({self.toplvl_name},\"{self.name}_\")\n"
        movecode += f"{self.toplvl_name}.add_ports({self.name}_ref.get_ports_list(),prefix=\"{self.name}_\")"
        return movecode
    
    @classmethod
    def test(cls):
        tests = list()
        tests.append(["mirror","top",(4.2,3.5)])
        tests.append(["ref","top",(9,5)])
        for testinst in tests:
            print(AbsoluteMove(*testinst).get_code())

class RelativeMove(GlayoutAction):
    """Move an existing Component relative to another component
    The following information is stored:
        name: str (name of component to move should match the place table)
        comp_id: str (name of the referenced component)
        direction: tuple[int,int] either 0,1,-1 representing x,y multiplier to separation (up/north/above, right/east, left/west, down/south/below)
        separation: str (variable or value but do not try to parse just send directly to code)
        strdirection: str to use in creating the comment for the move
    """
    move_index = int(0)

    #@validate_call
    def __init__(self, name_of_component_to_move: str, toplvl_name: str, relative_comp: str, direction: str, separation: str="maxmetalsep"):
        """Store all information neccessary to move a Component relative to another component
        Args:
            name_of_component_to_move (str): name of the component that will be moved
            relative_comp (str): name of component that is taken as a reference for the move
            direction (str): up/north/above, right/east, left/west, down/south/below
            separation (str): value, variable, or expression to use for separation between relative comp and comp we are moving
        """
        self.name = str(name_of_component_to_move)
        self.relative_comp = str(relative_comp)
        self.toplvl_name = toplvl_name
        # TODO: more sophisticated validation that can distinguish float from var for separation
        self.separation = str(separation) if separation is not None else "maxmetalsep"
        # parse direction
        direction = direction.lower().strip()
        self.strdirection = direction
        dirint = parse_direction(direction)
        if dirint==1:
            self.direction = (0,1)
        elif dirint==2:
            self.direction = (1,0)
        elif dirint==0:
            self.direction = (-1,0)
        elif dirint==3:
            self.direction = (0,-1)
        else:
            raise ValueError("invalid direction, move must be either up/north/above, right/east, left/west, or down/south/below")
    
    def get_code(self) -> str:
        movinfo = [self.separation*direc for direc in self.direction]
        l1 = f"# move {self.name} {self.strdirection} {self.relative_comp}"
        l2 = None
        if self.direction[0]==1:# move to the right
            l2 = f"{self.direction[0]}*({self.separation} + center_to_edge_distance({self.relative_comp}_ref,3) + center_to_edge_distance({self.name}_ref,1))"
            movfunc, cidx = "movex", 0
        if self.direction[0]==-1:# move to the left
            l2 = f"{self.direction[0]}*({self.separation} + center_to_edge_distance({self.relative_comp}_ref,1) + center_to_edge_distance({self.name}_ref,3))"
            movfunc, cidx  = "movex", 0
        if self.direction[1]==1:# move to the north
            l2 = f"{self.direction[1]}*({self.separation} + center_to_edge_distance({self.relative_comp}_ref,2) + center_to_edge_distance({self.name}_ref,4))"
            movfunc, cidx  = "movey", 1
        if self.direction[1]==-1:# move to the south
            l2 = f"{self.direction[1]}*({self.separation} + center_to_edge_distance({self.relative_comp}_ref,4) + center_to_edge_distance({self.name}_ref,2))"
            movfunc, cidx  = "movey", 1
        if l2 is None:
            raise ValueError("move must be either up/north/above, right/east, left/west, or down/south/below")
        l2 = f"relativemovcorrection_{str(self.move_index)} = " + l2
        l3 = movfunc + f"({self.name}_ref,destination=(relativemovcorrection_{str(self.move_index)} + {self.relative_comp}_ref.center[{cidx}]))"
        # update ports
        l4 = f"remove_ports_with_prefix({self.toplvl_name},\"{self.name}_\")"
        l5 = f"{self.toplvl_name}.add_ports({self.name}_ref.get_ports_list(),prefix=\"{self.name}_\")"
        return l1 + "\n" + l2 + "\n" + l3 + "\n" + l4+ "\n" + l5

    @classmethod
    def test(cls):
        raise NotImplementedError("testing RelativeMove has not yet been implemented")


class Route(GlayoutAction):
    """Route between two existing Ports (port1->port2)
    The following information is stored:
        port1: str (name of first port)
        port2: str (name of second port)
        route_type: handle (c,l,or straight route)
        params: dict paramters to pass to function
    """
    #@validate_call
    def __init__(self, toplvl_name: str, route_type: Callable, port1: str, port2: str, parameters: str, known_paramsORvars: list, compref: str=None):
        """Store all neccessary information to route between two ports
        Args:
            route_type (Callable): l,c, straight, or smart route
            port1 (str):
            port2 (str):
            parameters (str): user input parameters
            known_paramsORvars (list): names of currently defined parameters and variables
            compref (str, optional): if smart route this will be passed along
        """
        self.port1 = str(port1)
        self.port2 = str(port2)
        self.route_type = route_type
        # parse user params
        self.params = ParametersList(self.route_type).parse_user_params(known_paramsORvars, parameters)
        self.toplvl_name = toplvl_name
        self.compref = compref
    
    def get_code(self) -> str:
        port1s = f"{self.toplvl_name}.ports[\"{self.port1}\"]"
        port2s = f"{self.toplvl_name}.ports[\"{self.port2}\"]"
        if "smart" in self.route_type.__name__:
            return f"{self.toplvl_name} << {self.route_type.__name__}(pdk,{port1s},{port2s},{self.compref},{self.toplvl_name},**{str(self.params)})"
        return f"{self.toplvl_name} << {self.route_type.__name__}(pdk,{port1s},{port2s},**{str(self.params)})"
    
    @classmethod
    def test(cls):
        from glayout.flow.routing.c_route import c_route
        tests = list()
        tests.append([c_route,"name_of_port_1","name_of_port_2","cwidth of 4 width1=3"])
        for testinst in tests:
            print(Route(*testinst).get_code())



# Top level Code Relational table
class GlayoutCode(GlayoutAction):
    """Stores all needed information to create Glayout code in relation tables
    GlayoutActions require some general context provided by the GlayoutCode class"""
    
    HEAD_MARKER = "####\n# Compiled Glayout\n# Apache License\n# Version 2.0, January 2004\n# http://www.apache.org/licenses/"

    # each table is implemented as a list of dictionaries
    def __init__(self, toplvl_name: str):
        self.toplvl_name = toplvl_name.replace(" ","_").strip()
        # tables
        self.import_table = list()# list of ImportCell type
        self.parameter_table = list()# list of CreateCellParameter type
        self.bulk_action_table = list()# list of Route, AbsoluteMove, RelativeMove, PlaceCell, CreateWorkingVariable
        # initialize the import table to contain all primitives and routing primitives
        self.__known_generator_ids = list()# list of str
        self.__random_name_index = int(0)
        # primitives
        self.update_import_table(["nmos","nfet"],"nmos","glayout.flow.primitives.fet")
        self.update_import_table(["pmos","pfet"],"pmos","glayout.flow.primitives.fet")
        self.update_import_table(["guardring","tapring","welltap","well tap", "tap ring"],"tapring","glayout.flow.primitives.guardring")
        self.update_import_table(["mimcap"],"mimcap","glayout.flow.primitives.mimcap")
        self.update_import_table(["mimcap array","mimcaparray","mimcap_array"],"mimcap_array","glayout.flow.primitives.mimcap")
        self.update_import_table(["via","via stack","via_stack"],"via_stack","glayout.flow.primitives.via_gen")
        self.update_import_table(["via array","via_array"],"via_array","glayout.flow.primitives.via_gen")
        # general components and layout strategies
        two_nfet_interdigitized_aliases = ["interdigitized","interdigitated"]+list_cartesian_product(["interdigitized","interdigitated"],["nmos","nfet"],True)
        self.update_import_table(two_nfet_interdigitized_aliases,"two_nfet_interdigitized","glayout.flow.placement.two_transistor_interdigitized")
        generic_4T_interdigitzed_aliases = list_cartesian_product(list_cartesian_product(["four","4"], ["interdigitized","interdigitated"]),["fet","transistor"])
        self.update_import_table(generic_4T_interdigitzed_aliases, "generic_4T_interdigitzed", "glayout.flow.placement.four_transistor_interdigitized")
        two_pfet_interdigitized_aliases = list_cartesian_product(["interdigitized","interdigitated"],["pmos","pfet"],True)
        self.update_import_table(two_pfet_interdigitized_aliases,"two_pfet_interdigitized","glayout.flow.placement.two_transistor_interdigitized")
        self.update_import_table(["diff pair","diff_pair","differential pair","differential pairs","differential transistor"],"diff_pair_generic","glayout.flow.components.diff_pair")
        # import routing funcs
        self.update_import_table(["smart route","smart","smart_route"],"smart_route","glayout.flow.routing.smart_route")
        self.update_import_table(["L route","L_route","l route","l_route"],"L_route","glayout.flow.routing.L_route")
        self.update_import_table(["C route","C_route","c route","c_route"],"c_route","glayout.flow.routing.c_route")
        self.update_import_table(["straight route","straight_route"],"straight_route","glayout.flow.routing.straight_route")
        # general utils are hardcoded imports in the get_code method
        # add working variable max metal separation
        self.update_variable_table("maxmetalsep","pdk.util_max_metal_seperation()")
        self.update_variable_table("double_maxmetalsep","2*pdk.util_max_metal_seperation()")
        self.update_variable_table("triple_maxmetalsep","3*pdk.util_max_metal_seperation()")
        self.update_variable_table("quadruple_maxmetalsep","4*pdk.util_max_metal_seperation()")
        self.__placed_noname_objs = dict()
    
    def search_import_table(self, generator_id: str) -> Callable:
        """find a component with one identifier matching generator_id and return the function handle"""
        handle = None
        for generator in self.import_table:
            if any(generator_id==identifier for identifier in generator.identifiers):
                handle = generator.handle
        if handle is None:
            raise LookupError("the generator you were looking for was not found. Perhaps it uses a different alias?")
        return handle
    
    def update_import_table(self, component_identifiers: list[str], component_name: Optional[str]=None, module_path: Optional[Union[str, Path]]=None):
        """Add an ImportCell to the import_table, see ImportCell class"""
        if component_name is None or component_name=="" or component_name.isspace():
            component_name = f"comp_{self.__random_name_index}"
            self.__random_name_index += 1
        self.__known_generator_ids += component_identifiers
        self.import_table.append(ImportCell(component_identifiers, component_name, module_path))
    
    def update_parameter_table(
            self, 
            varname: str, 
            vartype: Optional[Literal[int, float]]=None, 
            defaultvalue: Optional[Union[int,float]]=None, 
            requirements: Optional[tuple[Literal[">",">=","<","<=","!=","=="],Union[float,int]]]=None
        ):
        """Add a CreateCellParameter to the parameter_table, see CreateCellParameter class"""
        self.parameter_table.append(CreateCellParameter(varname,vartype, defaultvalue, requirements))
    
    def update_variable_table(self, varname: str, input_expression: str):
        """Add a CreateWorkingVariable to the variable_table, see CreateWorkingVariable class"""
        self.bulk_action_table.append(CreateWorkingVariable(varname, input_expression))
    
    def update_place_table(self, generator_id: str, user_input_parameters: str, component_name: Optional[str]=None):
        """Add a PlaceCell to the place_table, see PlaceCell class
        Args:
            generator_id (str): a known name for cell factory to use
            component_name (str): name to assign to this component
            user_input_parameters (str): just the portion of user input containing the parameters for the component
            user input should be parameter name followed by value (optionally with a filler word e.g. [of, =, is])
            user input should contain NOTHING except parameters and values
        """
        if component_name is None:
            current_val = self.__placed_noname_objs.get(generator_id,int(0))
            self.__placed_noname_objs[generator_id] = current_val + 1
            component_name = generator_id + str(current_val)
        self.bulk_action_table.append(PlaceCell(self.toplvl_name, self.search_import_table(generator_id), component_name, user_input_parameters, self.get_params_and_vars()))
    
    def update_move_table(self, move_type: str, name_of_component_to_move: str, *args, **kwargs):
        """move_type can be absolute, relative"""
        move_type = move_type.lower().strip()
        if move_type=="absolute":
            self.bulk_action_table.append(AbsoluteMove(name_of_component_to_move,self.toplvl_name,*args,**kwargs))
        elif move_type=="relative":
            self.bulk_action_table.append(RelativeMove(name_of_component_to_move,self.toplvl_name,*args,**kwargs))
    
    def update_route_table(self, port1: str, port2: str, parameters: str, route_type: Optional[str]=None):
        # guess route type if not specified
        if route_type is None:
            route_type = "smart_route"
        else:
            # parse route type
            route_type = str(route_type).lower().strip()
            rpre = "L" if "l" in route_type else ("c" if "c" in route_type else ("straight" if "s" in route_type else None))
            rpre = "smart" if "sma" in route_type else rpre
            if rpre is None:
                rpre = "smart"
            route_type = rpre + "_route"
        # look for top comp ref this goes back to
        if route_type=="smart_route":
            for cellname in self.names_of_placed_cells():
                compref = cellname in port1 and cellname in port2
                if compref is not None:
                    compref = cellname + "_ref"
                    break
        else:
            compref = None
        self.bulk_action_table.append(Route(self.toplvl_name,self.search_import_table(route_type),port1,port2,parameters,self.get_params_and_vars(),compref))
    
    def get_params_and_vars(self) -> list[str]:
        """Get the names of parameters and variables.
        This method retrieves the names of parameters from the parameter table and the names of variables created
        by 'CreateWorkingVariable' actions from the bulk action table.
        Returns:
            list[str]: A list containing the names of parameters and variables.
        """
        params = [param.varname for param in self.parameter_table]
        variables = [variable.varname for variable in self.bulk_action_table if isinstance(variable,CreateWorkingVariable)]
        return params + variables
    
    def get_code(self) -> str:
        # time this NLP was compiled to python code
        current_time = datetime.datetime.now()
        compilation_head = self.HEAD_MARKER + f"\n# {current_time}\n\n"
        # produce all import code
        import_code = "from glayout.flow.pdk.mappedpdk import MappedPDK\n"
        import_code += "from gdsfactory import Component\n"
        import_code += "from glayout.flow.pdk.util.comp_utils import move, movex, movey, prec_ref_center, evaluate_bbox, center_to_edge_distance\n"
        import_code += "from glayout.flow.pdk.util.port_utils import remove_ports_with_prefix\n"
        for comp_import in self.import_table:
            import_code += comp_import.get_code() + "\n"
        # create function header
        function_head = f"def {self.toplvl_name}_cell(\n\tpdk: MappedPDK,\n"
        for param in self.parameter_table:
            function_head += param.get_code()+"\n"
        function_head += "):\n"
        # create variables, place, route, and move in the order that user supplied these directions
        function_body = "pdk.activate()\n"
        function_body += f"{self.toplvl_name} = Component(name=\"{self.toplvl_name}\")\n"
        for bulk_action in self.bulk_action_table:
            function_body += bulk_action.get_code()+"\n"
        function_body = "\n".join(["\t"+line for line in function_body.splitlines()])
        # footer
        footer = "\n\treturn "+self.toplvl_name+"\n"
        return compilation_head + import_code + "\n" + function_head + function_body + footer
    
    @classmethod
    def test(cls, test_all: bool=True):
        glayout_actions = get_all_derived_classes(GlayoutAction)
        glayout_actions.remove(cls)
        print("now testing individual Glayout Actions")
        for action in glayout_actions:
            print(f"\ntesting{action.__name__}")
            try:
                action.test()
            except Exception as e:
                print(e)
        print("\n\nnow testing GlayoutCode")
        codeobj = GlayoutCode("newtoplvl")
        codeobj.update_import_table(["stream","test_stream"],"test_stream","./deprecated/practice_stream.py")
        codeobj.update_parameter_table("width",float,5.5)
        codeobj.update_variable_table("int_width","int(width)")
        codeobj.update_place_table("nfet","mirror","width of 5 length is 2")
        codeobj.update_move_table("absolute","mirror",(3.3,3.3))
        codeobj.update_route_table("nfet_drain_west","nfet_source_west","","c route")
        print(codeobj.get_code())
    
    def names_of_placed_cells(self) -> list:
        """Get the names of cells that have been placed.
        Returns:
            list: A list containing the names of cells that have been placed.
        """
        names = list()
        for action in self.bulk_action_table:
            if isinstance(action,PlaceCell):
                names.append(action.name)
        return names

    def find_first_generator_id(self, sentence: str) -> str:
        """returns the longest generator id in the given sentence
        should only be used on sentences which for sure have a generator id in them
        """
        # look for genids found in the sentence
        genid_candidates = list()
        for genid in self.__known_generator_ids:
            if genid in sentence:
                genid_candidates.append(genid)
        # error checking
        if len(genid_candidates) == 0:
            print(self.__known_generator_ids)
            print(sentence)
            raise LookupError("no known generator id was found in the provided sentence")
        # prefer longer strings
        return max(genid_candidates, key=len)









####
# Compiled Glayout
# Apache License
# Version 2.0, January 2004
# http://www.apache.org/licenses/
# 2024-05-16 19:29:27.736502

from glayout.flow.pdk.mappedpdk import MappedPDK
from gdsfactory import Component
from glayout.flow.pdk.util.comp_utils import move, movex, movey, prec_ref_center, evaluate_bbox, center_to_edge_distance
from glayout.flow.pdk.util.port_utils import remove_ports_with_prefix
from glayout.flow.primitives.fet import nmos
from glayout.flow.primitives.fet import pmos
from glayout.flow.primitives.guardring import tapring
from glayout.flow.primitives.mimcap import mimcap
from glayout.flow.primitives.mimcap import mimcap_array
from glayout.flow.primitives.via_gen import via_stack
from glayout.flow.primitives.via_gen import via_array
from glayout.flow.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.flow.placement.four_transistor_interdigitized import generic_4T_interdigitzed
from glayout.flow.placement.two_transistor_interdigitized import two_pfet_interdigitized
from glayout.flow.components.diff_pair import diff_pair_generic
from glayout.flow.routing.smart_route import smart_route
from glayout.flow.routing.L_route import L_route
from glayout.flow.routing.c_route import c_route
from glayout.flow.routing.straight_route import straight_route
from glayout.flow.components.diff_pair import diff_pair
from glayout.flow.placement.four_transistor_interdigitized import generic_4T_interdigitzed
from glayout.flow.components.current_mirror import current_mirror

def opamp_cell(
	pdk: MappedPDK,
	numcols: int, 
	load_pfet_width: float, 
	load_pfet_fingers: int, 
	load_pfet_multipliers: int, 
	load_curr_source_fingers: int, 
	load_curr_source_width_ref: float, 
	load_curr_source_width_mirr: float, 
	mim_cap_size: tuple, 
	mim_cap_rows: int, 
):
	pdk.activate()
	opamp = Component(name="opamp")
	maxmetalsep = pdk.util_max_metal_seperation()
	double_maxmetalsep = 2*pdk.util_max_metal_seperation()
	triple_maxmetalsep = 3*pdk.util_max_metal_seperation()
	quadruple_maxmetalsep = 4*pdk.util_max_metal_seperation()
	# placing mydiffload centered at the origin
	mydiffload = generic_4T_interdigitzed(pdk,**{'numcols': numcols, 'top_kwargs': { "tie_layers" : ( "met3" , "met2" ) }, 'top_row_device': "pfet", 'bottom_row_device': "pfet"})
	mydiffload_ref = prec_ref_center(mydiffload)
	opamp.add(mydiffload_ref)
	opamp.add_ports(mydiffload_ref.get_ports_list(),prefix="mydiffload_")
	# placing mydiff centered at the origin
	mydiff = diff_pair(pdk,**{'width': 5, 'fingers': numcols, 'dummy': True, 'substrate_tap': True})
	mydiff_ref = prec_ref_center(mydiff)
	opamp.add(mydiff_ref)
	opamp.add_ports(mydiff_ref.get_ports_list(),prefix="mydiff_")
	# move mydiffload north mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,2) + center_to_edge_distance(mydiffload_ref,4))
	movey(mydiffload_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"mydiffload_")
	opamp.add_ports(mydiffload_ref.get_ports_list(),prefix="mydiffload_")
	opamp << smart_route(pdk,opamp.ports["mydiffload_top_A_source_W"],opamp.ports["mydiffload_top_B_source_W"],mydiffload_ref,opamp,**{})
	opamp << c_route(pdk,opamp.ports["mydiffload_top_A_gate_W"],opamp.ports["mydiffload_top_B_gate_W"],**{'extension': 2})
	opamp << c_route(pdk,opamp.ports["mydiffload_bottom_A_gate_W"],opamp.ports["mydiffload_bottom_B_gate_W"],**{'extension': 2})
	opamp << smart_route(pdk,opamp.ports["mydiffload_bottom_A_gate_E"],opamp.ports["mydiffload_top_A_gate_E"],mydiffload_ref,opamp,**{})
	opamp << smart_route(pdk,opamp.ports["mydiffload_bottom_A_gate_E"],opamp.ports["mydiffload_top_A_gate_E"],mydiffload_ref,opamp,**{})
	opamp << smart_route(pdk,opamp.ports["mydiffload_top_A_drain_W"],opamp.ports["mydiffload_bottom_A_source_W"],mydiffload_ref,opamp,**{})
	opamp << c_route(pdk,opamp.ports["mydiffload_top_B_drain_E"],opamp.ports["mydiffload_bottom_B_source_E"],**{'extension': 2})
	opamp << smart_route(pdk,opamp.ports["mydiffload_bottom_B_drain_W"],opamp.ports["mydiffload_bottom_B_gate_W"],mydiffload_ref,opamp,**{})
	opamp << c_route(pdk,opamp.ports["mydiffload_top_A_drain_W"],opamp.ports["mydiff_bl_multiplier_0_drain_W"],**{'extension': 4})
	opamp << c_route(pdk,opamp.ports["mydiffload_top_B_drain_E"],opamp.ports["mydiff_br_multiplier_0_drain_E"],**{'extension': 4})
	# placing mycurrmirror centered at the origin
	mycurrmirror = current_mirror(pdk,**{'numcols': numcols, 'device': "nfet", 'with_dummy': False})
	mycurrmirror_ref = prec_ref_center(mycurrmirror)
	opamp.add(mycurrmirror_ref)
	opamp.add_ports(mycurrmirror_ref.get_ports_list(),prefix="mycurrmirror_")
	# move mycurrmirror south mydiff
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mydiff_ref,4) + center_to_edge_distance(mycurrmirror_ref,2))
	movey(mycurrmirror_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"mycurrmirror_")
	opamp.add_ports(mycurrmirror_ref.get_ports_list(),prefix="mycurrmirror_")
	opamp << c_route(pdk,opamp.ports["mydiff_bl_multiplier_0_source_W"],opamp.ports["mycurrmirror_fet_B_0_source_W"],**{'extension': 3})
	# placing load_pfet_left centered at the origin
	load_pfet_left = pmos(pdk,**{'width': load_pfet_width, 'fingers': load_pfet_fingers, 'multipliers': load_pfet_multipliers})
	load_pfet_left_ref = prec_ref_center(load_pfet_left)
	opamp.add(load_pfet_left_ref)
	opamp.add_ports(load_pfet_left_ref.get_ports_list(),prefix="load_pfet_left_")
	# move load_pfet_left left mydiff
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mydiff_ref,1) + center_to_edge_distance(load_pfet_left_ref,3))
	movex(load_pfet_left_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_pfet_left_")
	opamp.add_ports(load_pfet_left_ref.get_ports_list(),prefix="load_pfet_left_")
	# move load_pfet_left north mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,2) + center_to_edge_distance(load_pfet_left_ref,4))
	movey(load_pfet_left_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"load_pfet_left_")
	opamp.add_ports(load_pfet_left_ref.get_ports_list(),prefix="load_pfet_left_")
	# move load_pfet_left left mydiffload
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mydiffload_ref,1) + center_to_edge_distance(load_pfet_left_ref,3))
	movex(load_pfet_left_ref,destination=(relativemovcorrection_0 + mydiffload_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_pfet_left_")
	opamp.add_ports(load_pfet_left_ref.get_ports_list(),prefix="load_pfet_left_")
	# placing load_pfet_right centered at the origin
	load_pfet_right = pmos(pdk,**{'width': load_pfet_width, 'fingers': load_pfet_fingers, 'multipliers': load_pfet_multipliers})
	load_pfet_right_ref = prec_ref_center(load_pfet_right)
	opamp.add(load_pfet_right_ref)
	opamp.add_ports(load_pfet_right_ref.get_ports_list(),prefix="load_pfet_right_")
	# move load_pfet_right right mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,3) + center_to_edge_distance(load_pfet_right_ref,1))
	movex(load_pfet_right_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_pfet_right_")
	opamp.add_ports(load_pfet_right_ref.get_ports_list(),prefix="load_pfet_right_")
	# move load_pfet_right north mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,2) + center_to_edge_distance(load_pfet_right_ref,4))
	movey(load_pfet_right_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"load_pfet_right_")
	opamp.add_ports(load_pfet_right_ref.get_ports_list(),prefix="load_pfet_right_")
	# move load_pfet_right right mydiffload
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiffload_ref,3) + center_to_edge_distance(load_pfet_right_ref,1))
	movex(load_pfet_right_ref,destination=(relativemovcorrection_0 + mydiffload_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_pfet_right_")
	opamp.add_ports(load_pfet_right_ref.get_ports_list(),prefix="load_pfet_right_")
	opamp << c_route(pdk,opamp.ports["load_pfet_right_multiplier_0_gate_con_N"],opamp.ports["load_pfet_left_multiplier_0_gate_con_N"],**{'extension': 18, 'width1': 0.8, 'width2': 0.8})
	opamp << c_route(pdk,opamp.ports["load_pfet_right_multiplier_0_drain_con_N"],opamp.ports["load_pfet_left_multiplier_0_drain_con_N"],**{'extension': 8, 'width1': 0.8, 'width2': 0.8})
	opamp << c_route(pdk,opamp.ports["load_pfet_right_multiplier_0_source_con_N"],opamp.ports["load_pfet_left_multiplier_0_source_con_N"],**{'extension': 5.5, 'viaoffset': ( True , False ), 'fullbottom': True, 'width1': 0.8, 'width2': 0.8})
	# placing load_curr_source_ref centered at the origin
	load_curr_source_ref = nmos(pdk,**{'width': load_curr_source_width_ref, 'fingers': load_curr_source_fingers, 'multipliers': 2})
	load_curr_source_ref_ref = prec_ref_center(load_curr_source_ref)
	opamp.add(load_curr_source_ref_ref)
	opamp.add_ports(load_curr_source_ref_ref.get_ports_list(),prefix="load_curr_source_ref_")
	# move load_curr_source left mydiff
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mydiff_ref,1) + center_to_edge_distance(load_curr_source_ref,3))
	movex(load_curr_source_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_curr_source_")
	opamp.add_ports(load_curr_source_ref.get_ports_list(),prefix="load_curr_source_")
	# move load_curr_source_ref south mydiff
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mydiff_ref,4) + center_to_edge_distance(load_curr_source_ref_ref,2))
	movey(load_curr_source_ref_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"load_curr_source_ref_")
	opamp.add_ports(load_curr_source_ref_ref.get_ports_list(),prefix="load_curr_source_ref_")
	# move load_curr_source_ref left mycurrmirror
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mycurrmirror_ref,1) + center_to_edge_distance(load_curr_source_ref_ref,3))
	movex(load_curr_source_ref_ref,destination=(relativemovcorrection_0 + mycurrmirror_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_curr_source_ref_")
	opamp.add_ports(load_curr_source_ref_ref.get_ports_list(),prefix="load_curr_source_ref_")
	# placing load_curr_source_mirr centered at the origin
	load_curr_source_mirr = nmos(pdk,**{'width': load_curr_source_width_mirr, 'fingers': load_curr_source_fingers, 'multipliers': 2})
	load_curr_source_mirr_ref = prec_ref_center(load_curr_source_mirr)
	opamp.add(load_curr_source_mirr_ref)
	opamp.add_ports(load_curr_source_mirr_ref.get_ports_list(),prefix="load_curr_source_mirr_")
	# move load_curr_source_mirr right mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,3) + center_to_edge_distance(load_curr_source_mirr_ref,1))
	movex(load_curr_source_mirr_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_curr_source_mirr_")
	opamp.add_ports(load_curr_source_mirr_ref.get_ports_list(),prefix="load_curr_source_mirr_")
	# move load_curr_source_mirr south mydiff
	relativemovcorrection_0 = -1*(maxmetalsep + center_to_edge_distance(mydiff_ref,4) + center_to_edge_distance(load_curr_source_mirr_ref,2))
	movey(load_curr_source_mirr_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"load_curr_source_mirr_")
	opamp.add_ports(load_curr_source_mirr_ref.get_ports_list(),prefix="load_curr_source_mirr_")
	# move load_curr_source_mirr right mycurrmirror
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mycurrmirror_ref,3) + center_to_edge_distance(load_curr_source_mirr_ref,1))
	movex(load_curr_source_mirr_ref,destination=(relativemovcorrection_0 + mycurrmirror_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_curr_source_mirr_")
	opamp.add_ports(load_curr_source_mirr_ref.get_ports_list(),prefix="load_curr_source_mirr_")
	opamp << c_route(pdk,opamp.ports["load_curr_source_ref_multiplier_0_gate_con_S"],opamp.ports["load_curr_source_mirr_multiplier_0_gate_con_S"],**{'extension': 3, 'width1': 0.8, 'width2': 0.8})
	opamp << c_route(pdk,opamp.ports["load_curr_source_ref_multiplier_0_drain_con_S"],opamp.ports["load_curr_source_mirr_multiplier_0_drain_con_S"],**{'extension': 12, 'viaoffset': ( False , True ), 'width1': 0.8, 'width2': 0.8})
	opamp << c_route(pdk,opamp.ports["load_curr_source_ref_multiplier_0_source_con_S"],opamp.ports["load_curr_source_mirr_multiplier_0_source_con_S"],**{'extension': 14, 'width1': 0.8, 'width2': 0.8})
	opamp << c_route(pdk,opamp.ports["load_curr_source_ref_multiplier_0_gate_E"],opamp.ports["load_curr_source_ref_multiplier_0_drain_E"],**{'extension': 3})
	opamp << L_route(pdk,opamp.ports["mydiff_tr_multiplier_0_drain_W"],opamp.ports["load_pfet_right_multiplier_0_gate_con_S"],**{})
	opamp << L_route(pdk,opamp.ports["load_pfet_right_multiplier_0_drain_con_S"],opamp.ports["load_curr_source_mirr_multiplier_0_drain_W"],**{})
	# placing load_miller_cap centered at the origin
	load_miller_cap = mimcap_array(pdk,**{'rows': mim_cap_rows, 'columns': 2, 'size': mim_cap_size})
	load_miller_cap_ref = prec_ref_center(load_miller_cap)
	opamp.add(load_miller_cap_ref)
	opamp.add_ports(load_miller_cap_ref.get_ports_list(),prefix="load_miller_cap_")
	# move load_miller_cap right mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,3) + center_to_edge_distance(load_miller_cap_ref,1))
	movex(load_miller_cap_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_miller_cap_")
	opamp.add_ports(load_miller_cap_ref.get_ports_list(),prefix="load_miller_cap_")
	# move load_miller_cap north mydiff
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(mydiff_ref,2) + center_to_edge_distance(load_miller_cap_ref,4))
	movey(load_miller_cap_ref,destination=(relativemovcorrection_0 + mydiff_ref.center[1]))
	remove_ports_with_prefix(opamp,"load_miller_cap_")
	opamp.add_ports(load_miller_cap_ref.get_ports_list(),prefix="load_miller_cap_")
	# move load_miller_cap right load_pfet_right
	relativemovcorrection_0 = 1*(maxmetalsep + center_to_edge_distance(load_pfet_right_ref,3) + center_to_edge_distance(load_miller_cap_ref,1))
	movex(load_miller_cap_ref,destination=(relativemovcorrection_0 + load_pfet_right_ref.center[0]))
	remove_ports_with_prefix(opamp,"load_miller_cap_")
	opamp.add_ports(load_miller_cap_ref.get_ports_list(),prefix="load_miller_cap_")
	opamp << c_route(pdk,opamp.ports["mydiffload_top_B_0_drain_N"],opamp.ports["load_miller_cap_row0_col0_top_met_N"],**{'extension': 15})
	opamp << c_route(pdk,opamp.ports["load_pfet_right_multiplier_0_drain_con_S"],opamp.ports["load_miller_cap_row0_col0_top_met_S"],**{'width1': 1.2, 'width2': 1.2})
	return opamp



from setuptools import setup, find_packages


setup(
    name="glayout",
    version="0.0.6",
    author="Ali Hammoud, Harsh Khandeparkar, Vijay Shankar, Chetanya Goyal, Sakib Pathen, Arlene Dai, Ryan Wans, Mehdi Saligane",
    author_email="alibilal@umich.edu, Harsh, vijayshankar.renganathan@analog.com, Chetanya, spathen@umich.edu, arlendai@umich.edu, Ryan, mehdi@umich.edu",
    description="A human language to analog layout API with support for different technologies.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/idea-fasoc/OpenFASOC/tree/main/openfasoc/generators/glayout",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "gdsfactory==7.7.0",
        "prettyprint",
        "prettyprinttree",
        "nltk"
    ],
    extras_require={
        "llm": [
            "torch",
            "transformers",
            "langchain",
            "langchain_community",
            "chromadb",
            "ollama",
            "unstructured",
            "unstructured[md]",
            "sentence-transformers",
            "peft",
            "accelerate",
            "bitsandbytes",
            "safetensors",
            "datasets",
            "auto-gptq",
            "optimum",
            "trl",
            "langchain_huggingface",
            "tensorboard"
        ]
    },
    entry_points={
        "console_scripts": [
            # "glayout = something:run"# Define any command line scripts here
        ],
    },
)



import nltk


def check_and_download_nltk_data(data_name):
    """
    Check if an NLTK dataset is downloaded, and download it if it is not.

    Args:
        data_name (str): The name of the NLTK dataset to check and download.
    """
    try:
        nltk.data.find(f"tokenizers/{data_name}")
    except LookupError:
        print(f"{data_name} is not downloaded. Downloading now...")
        nltk.download(data_name)
        print(f"{data_name} has been downloaded.")


check_and_download_nltk_data("punkt")



from glayout.flow.placement.two_transistor_interdigitized import two_nfet_interdigitized, two_pfet_interdigitized
from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.routing.c_route import c_route
from glayout.flow.routing.L_route import L_route
from glayout.flow.routing.straight_route import straight_route
from glayout.flow.pdk.sky130_mapped import sky130_mapped_pdk as sky130
from glayout.flow.primitives.fet import nmos, pmos
from glayout.flow.primitives.guardring import tapring
from glayout.flow.pdk.util.port_utils import add_ports_perimeter	
from gdsfactory.component import Component
from typing import Optional 

#@cell
def current_mirror(
    pdk: MappedPDK, 
    numcols: int = 3,
    device: str = 'nfet',
    with_dummy: Optional[bool] = True,
    with_substrate_tap: Optional[bool] = False,
    with_tie: Optional[bool] = True,
    tie_layers: tuple[str,str]=("met2","met1"),
    **kwargs
) -> Component:
	"""An instantiable current mirror that returns a Component object. The current mirror is a two transistor interdigitized structure with a shorted source and gate. It can be instantiated with either nmos or pmos devices. It can also be instantiated with a dummy device, a substrate tap, and a tie layer, and is centered at the origin. Transistor A acts as the reference and Transistor B acts as the mirror fet

	Args:
		pdk (MappedPDK): the process design kit to use
		numcols (int): number of columns of the interdigitized fets
		device (str): nfet or pfet (can only interdigitize one at a time with this option)
		with_dummy (bool): True places dummies on either side of the interdigitized fets
		with_substrate_tap (bool): boolean to decide whether to place a substrate tapring
		with_tie (bool): boolean to decide whether to place a tapring for tielayer
		tie_layers (tuple[str,str], optional): the layers to use for the tie. Defaults to ("met2","met1").
		**kwargs: The keyword arguments are passed to the two_nfet_interdigitized or two_pfet_interdigitized functions and need to be valid arguments that can be accepted by the multiplier function

	Returns:
		Component: a current mirror component object
	"""
	top_level = Component("current mirror")
	if device in ['nmos', 'nfet']:
		interdigitized_fets = two_nfet_interdigitized(
			pdk, 
			numcols=numcols, 
			dummy=with_dummy, 
			with_substrate_tap=False, 
			with_tie=False, 
			**kwargs
		)
	elif device in ['pmos', 'pfet']:
		interdigitized_fets = two_pfet_interdigitized(
			pdk, 
			numcols=numcols, 
			dummy=with_dummy, 
			with_substrate_tap=False, 
			with_tie=False, 
			**kwargs
		)
	top_level.add_ports(interdigitized_fets.get_ports_list(), prefix="fet_")
	maxmet_sep = pdk.util_max_metal_seperation()
	# short source of the fets
	source_short = top_level << c_route(pdk, interdigitized_fets.ports['A_source_E'], interdigitized_fets.ports['B_source_E'], extension=3*maxmet_sep, viaoffset=False)
	# short gates of the fets
	gate_short = top_level << c_route(pdk, interdigitized_fets.ports['A_gate_W'], interdigitized_fets.ports['B_gate_W'], extension=3*maxmet_sep, viaoffset=False)
	# short gate and drain of one of the reference 
	top_level << L_route(pdk, interdigitized_fets.ports['A_drain_W'], gate_short.ports['con_N'], viaoffset=False, fullbottom=False)
	
	# add the tie layer
	if with_tie:
		tap_sep = max(
            pdk.util_max_metal_seperation(),
            pdk.get_grule("active_diff", "active_tap")["min_separation"],
        )
		tap_sep += pdk.get_grule("p+s/d", "active_tap")["min_enclosure"]
		tap_encloses = (
		2 * (tap_sep + interdigitized_fets.xmax),
		2 * (tap_sep + interdigitized_fets.ymax),
		)
		tie_ref = top_level << tapring(pdk, enclosed_rectangle = tap_encloses, sdlayer = "p+s/d", horizontal_glayer = tie_layers[0], vertical_glayer = tie_layers[1])
		top_level.add_ports(tie_ref.get_ports_list(), prefix="welltie_")
		try:
			top_level << straight_route(pdk, top_level.ports["A_0_dummy_L_gsdcon_top_met_W"],top_level.ports["welltie_W_top_met_W"],glayer2="met1")
		except KeyError:
			pass
		try:
			end_col = numcols - 1
			port1 = f'B_{end_col}_dummy_R_gdscon_top_met_E'
			top_level << straight_route(pdk, top_level.ports[port1], top_level.ports["welltie_E_top_met_E"], glayer2="met1")
		except KeyError:
			pass
	
	# add a pwell 
	top_level.add_padding(layers = (pdk.get_glayer("pwell"),), default = pdk.get_grule("pwell", "active_tap")["min_enclosure"], )
	top_level = add_ports_perimeter(top_level, layer = pdk.get_glayer("pwell"), prefix="well_")
 
	# add the substrate tap if specified
	if with_substrate_tap:
		subtap_sep = pdk.get_grule("dnwell", "active_tap")["min_separation"]
		subtap_enclosure = (
			2.5 * (subtap_sep + interdigitized_fets.xmax),
			2.5 * (subtap_sep + interdigitized_fets.ymax),
		)
		subtap_ring = top_level << tapring(pdk, enclosed_rectangle = subtap_enclosure, sdlayer = "p+s/d", horizontal_glayer = "met2", vertical_glayer = "met1")
		top_level.add_ports(subtap_ring.get_ports_list(), prefix="substrate_tap_")
  
	top_level.add_ports(source_short.get_ports_list(), prefix='purposegndports')
	top_level << interdigitized_fets
	return top_level


from pathlib import Path
import json
from typing import Union
import time
import argparse

from glayout.llm.manage_data import (
    load_preprocessed_pretokenized_data,
    unify_prompt_and_add_context_to_data,
    get_glayout_context,
    get_prompt_from_template,
    load_preprocessed_data_in_messages_format,
    load_all_labeled_syntax_data_json,
)

import torch
from peft import (
    get_peft_config,
    get_peft_model,
    LoraConfig,
    prepare_model_for_kbit_training,
    AutoPeftModelForCausalLM,
)
from datasets import Dataset
from auto_gptq import AutoGPTQForCausalLM

from transformers import (
    AutoModelForCausalLM,
    AutoModel,
    AutoTokenizer,
    TrainingArguments,
    BitsAndBytesConfig,
)
import transformers
from trl import DataCollatorForCompletionOnlyLM, SFTTrainer


def get_huggingface_token():
    """Parse the command-line arguments to retrieve the Hugging Face access token.
    This function uses argparse to handle command-line arguments and specifically looks for an 
    access token required to download models and tokenizers from Hugging Face. If the token is 
    not provided, it raises an EnvironmentError with instructions on how to obtain one.
    Returns:
        str: The Hugging Face access token.
    Raises:
        EnvironmentError: If the access token is not provided in the command-line arguments.
    """
    parser = argparse.ArgumentParser(description="Manage, interact, and run the Glayout LLM")
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        help="Specify the access token you are using to download the model and tokenizer from huggingface",
    )
    args = parser.parse_args()
    if args.token is None:
        errstring = "To download models from huggingface you need a hugging face account and an access token"
        errstring += "\nYou can create a hugging face account here: https://huggingface.co/join\n"
        errstring += "Once you have an account and sign in, you can create an access token (need read access) here:\n"
        errstring += "https://huggingface.co/settings/tokens\n"
        errstring += "pass the access token in the command line with the option --token=[insert token here]"
        raise EnvironmentError(errstring)
    return args.token
#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
accesstoken = get_huggingface_token()


# returns model, tokenizer
def load_model_and_tokenizer(device: str, lora: bool = True) -> tuple:
    """Downloads or restores model and tokenizer
    converts the model to half precision
    moves tokenizer and model to the specified device

    Args:
        device (str): move model to device (tokenizer always runs on CPU)
                      (e.g., 'cpu', 'cuda').

    Returns:
        tuple: first element is model and second is tokenizer.

    Raises:
        ValueError: If there is an error in loading the model or tokenizer.
        RuntimeError: If there is an error moving the model to the specified device.
    """
    qlora = True
    # load modela
    # modelname = "mistralai/Mistral-7B-v0.1"
    modelname = "mistralai/Mistral-7B-Instruct-v0.3"
    if not qlora:
        model = AutoModelForCausalLM.from_pretrained(modelname, token=accesstoken)
    else:
        # modelname = "TheBloke/Mistral-7B-Instruct-v0.2-GPTQ"
        # model = AutoModelForCausalLM.from_pretrained(modelname, token=accesstoken, device_map="auto", load_in_8bit=True)
        model = AutoModelForCausalLM.from_pretrained(
            modelname,
            token=accesstoken,
            quantization_config=BitsAndBytesConfig(load_in_8bit=True)
        )
        # model = AutoModelForCausalLM.from_pretrained(modelname, token=accesstoken, device_map="auto", trust_remote_code=False, revision="main")
        model.train()
        model.gradient_checkpointing_enable()
        model = prepare_model_for_kbit_training(model)
    tokenizer = AutoTokenizer.from_pretrained(
        modelname, use_fast=True, token=accesstoken
    )
    # configure lora
    if lora:
        peft_config = LoraConfig(
            task_type="CAUSAL_LM",
            r=8,
            lora_alpha=16,
            lora_dropout=0.05,
            bias="none",
            target_modules=["q_proj", "k_proj", "v_proj", "o_proj"],
        )
        model = get_peft_model(model, peft_config)
        model.print_trainable_parameters()
    if not qlora:  # the model loaded by qlora is prequantized
        model.half()
        model.to(device)
    tokenizer.add_special_tokens({'pad_token': '[PAD]'})
    return model, tokenizer


def run_llm_normal(
    model, tokenizer, device: str, prompt: str, max_new_tokens: int = 1000
) -> str:
    """Generate a text completion for a given prompt using a provided language model.
    Args:
        model: The language model to use, should be compatible with huggingfaceinterface
        device (str): The device where the model is currently located
        prompt (str): The initial text to prompt the language model with.
        max_new_tokens (int, optional): maximum number of new tokens to generate. Defaults to 500.
    Returns:
        str: The text generated by the language model as a continuation of the prompt.
    """
    model.eval()
    inputs = tokenizer.apply_chat_template(
        [{"role": "user", "content": prompt}],
        tokenize=True,
        add_generation_prompt=True,
        return_tensors="pt",
    )
    inputs = inputs.to(device)
    outputs = model.generate(input_ids=inputs, max_new_tokens=max_new_tokens)
    return tokenizer.decode(outputs[0], skip_special_tokens=True)


def train(model, tokenizer, data, qlora: bool = True):
    if not qlora:
        raise NotImplementedError("currently only support qlora")
    # model.train()
    # hyperparameters
    lr = 1e-4
    batch_size = 1  # 2 #4
    num_epochs = 2
    # define training arguments
    output_dir = Path(__file__).resolve().parent / "glayout_llm_checkpoints"
    training_args = TrainingArguments(
        output_dir=str(output_dir),
        learning_rate=lr,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        num_train_epochs=num_epochs,
        weight_decay=0.01,
        logging_strategy="epoch",
        evaluation_strategy="epoch",
        save_strategy="epoch",
        load_best_model_at_end=True,
        gradient_accumulation_steps=1,
        warmup_steps=0,
        bf16=True,
        optim="paged_adamw_8bit",
    )
    # inlcude in the prompt do not repeat the context
    # try to see Mistral 7b docs if there is another label
    # try to only train on the response
    # experiment with these results include prompts or not
    # check the context length for Mistral
    # code distral, try to directly create from the python code.
    data_collator = transformers.DataCollatorForLanguageModeling(tokenizer, mlm=False)
    # data_collator = DataCollatorForCompletionOnlyLM(response_template="[/INST]",tokenizer=tokenizer,mlm=False)
    # configure trainer
    trainer = transformers.Trainer(
        model=model,
        train_dataset=data["train"],
        eval_dataset=data["evaluation"],
        args=training_args,
        data_collator=data_collator,
    )
    # train model
    model.config.use_cache = False  # silence warnings
    trainer.train()
    model.config.use_cache = True  # reenable warnings
    # model.to("cuda")
    model.save_pretrained(output_dir / "checkpoint-bestperf")
    model.eval()
    return model


def run_full_training() -> tuple:
    """returns model (and tokenizer) resulting from training LLM
    Returns:
        tuple: model, tokenizer
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model, tokenizer = load_model_and_tokenizer(device)
    # load fine tuning data
    data = load_preprocessed_pretokenized_data(tokenizer)
    return train(model, tokenizer, data), tokenizer


def run_full_SFT_training() -> tuple:
    # load model, tokenizer
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model, tokenizer = load_model_and_tokenizer(device)
    # load data
    data = load_preprocessed_data_in_messages_format()
    # train
    # hyperparameters
    lr = 5e-5
    batch_size = 1  # 2 #4
    num_epochs = 4
    # define training arguments
    output_dir = Path(__file__).resolve().parent / "glayout_llm_checkpoints"
    training_args = TrainingArguments(
        output_dir=str(output_dir),
        learning_rate=lr,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        num_train_epochs=num_epochs,
        weight_decay=0.01,
        logging_strategy="epoch",
        evaluation_strategy="epoch",
        save_strategy="epoch",
        load_best_model_at_end=True,
        gradient_accumulation_steps=1,
        warmup_steps=1,
        bf16=True,
        optim="paged_adamw_8bit",
    )
    #training_args = TrainingArguments(output_dir=str(output_dir))
    data_collator = DataCollatorForCompletionOnlyLM(response_template="[/INST]",instruction_template="[INST]",tokenizer=tokenizer,mlm=False)
    #import pdb; pdb.set_trace()
    trainer = SFTTrainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        train_dataset=data["train"],
        eval_dataset=data["evaluation"],
        max_seq_length=4096,
        data_collator=data_collator
    )# add context to all glayout prompts
    trainer.train()
    model.save_pretrained(output_dir / "checkpoint-bestperf")
    model.eval()
    return model, tokenizer


class GlayoutLLMSessionHandler:
    def __init__(self):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        # look for an existing model
        base_path = Path(__file__).resolve().parent
        checkpoint_dirs = list(base_path.glob("**/*checkpoint-bestperf*"))
        checkpoint_dir = None
        if len(checkpoint_dirs) > 0 and checkpoint_dirs[-1].is_dir():
            checkpoint_dir = checkpoint_dirs[-1]
        # if no existing model then run training
        if checkpoint_dir:
            print(f"Found checkpoint directory: {checkpoint_dir}")
            model, tokenizer = self.load_model_from_checkpoint(checkpoint_dir)
            # model.to(self.device)
            print("Model and tokenizer loaded successfully.")
        else:
            # model, tokenizer = run_full_training()
            model, tokenizer = run_full_SFT_training()
        # set self attributes
        #self.promptexamples = "the following are several labeled examples of converting prompts to strict syntax.\n"
        #promptexamples = load_all_labeled_syntax_data_json()
        #for prompt, result in promptexamples[::25]:
        #    self.promptexamples += prompt + "\n" + result + "\n\n"
        self.model = model
        self.tokenizer = tokenizer
        self.chat_history = []
        self.chat_history.append({"role": "user", "content": get_glayout_context()})
        self.chat_history.append({"role": "assistant", "content": RESPONSE})
        #print(self.generate(self.promptexamples, clear=False))
        #print(self.generate(user_input="summarize the following:\n" + get_glayout_context(), clear=False))

    def load_model_from_checkpoint(self, checkpoint_dir):
        # helper function
        def get_base_model_name_or_path(file_path: Union[str, Path]) -> str:
            file_path = Path(file_path)
            with file_path.open("r") as file:
                data = json.load(file)
            return data.get("base_model_name_or_path")

        # load model
        model = AutoPeftModelForCausalLM.from_pretrained(
            checkpoint_dir, device_map=self.device
        )
        model_id = get_base_model_name_or_path(checkpoint_dir / "adapter_config.json")
        # basemodel = AutoModelForCausalLM.from_pretrained(model_id, device_map=self.device)
        # model = AutoGPTQForCausalLM.from_quantized(checkpoint_dir)
        # model = model.merge_and_unload()
        # load tokenizer
        tokenizer = AutoTokenizer.from_pretrained(model_id, use_fast=True, token=accesstoken)
        return model, tokenizer

    def generate(self, user_input: str, clear: bool = False) -> str:
        """provide LLM output from user input
        by default will keep appending to the previous prompts in a conversation.
        Args:
            user_input (str): general user prompt
            clear (bool, Optional): reset the chat history. Default False
        Returns:
            str: strictsyntax output
        """
        self.model.eval()
        user_input = f"Glayout strictsyntax is a electronic circuit layout command language. Convert the following prompt to Glayout strictsyntax:\n{user_input}"
        if clear:
            self.chat_history = []
            self.generate(
                user_input="summarize the following:\n" + get_glayout_context(),
                clear=False,
            )
        self.chat_history.append({"role": "user", "content": user_input})
        inputs = self.tokenizer.apply_chat_template(
            self.chat_history,
            tokenize=True,
            add_generation_prompt=True,
            return_tensors="pt",
        )
        inputs = inputs.to(self.device)
        outputs = self.model.generate(input_ids=inputs, max_new_tokens=4096)
        response = self.tokenizer.decode(
            outputs[0][len(inputs[0]) : -1], skip_special_tokens=False
        )
        self.chat_history.append({"role": "assistant", "content": response})
        return response
        # prompt = unify_prompt_and_add_context_to_data(self.tokenizer, input_list, no_label=True)[0]
        # return run_llm_normal(model=self.model, tokenizer=self.tokenizer, device=self.device, prompt=prompt)

    def __call__(self, user_input: str) -> str:
        return self.generate(user_input=user_input)


RESPONSE = """Example Syntax:
Importing: import CrossCoupledInverters
Creating Parameters: create a float parameter called device_width
Placing Components: place a nmos called m1 with width 1.0, length 2.0, fingers 2
Moving Components: move m1 below m2
Routing: route between m1_source_E and m2_source_E using smart_route
This structured approach ensures clarity and modularity, making it easier to design complex analog circuits efficiently."""
